import numpy as np
import numpy.linalg as la
import os as os

import topology.bounding.sphere as sphere
import topology.util.math as mutil

def ismax(vertices, centre, radius, basis, relative, max_radius):
    """ Returns true if the vertices indexed in relative is a
        simplex of maximal dimension in vertices

    Args:
        vertices: entire vertex list
        centre: centre of sphere
        radius: radius of sphere
        basis: indices of vertices that lie on the sphere
        relative: indices of vertices in simplex
        max_radius: max radius of enclosing sphere

    Returns:
        ismaxcell: true/false
    """
    relative_vertices = vertices[relative-1, :]

    dimension = relative_vertices.shape[1]
    n = relative_vertices.shape[0]

    # translate vertices so that centre is at origin
    ones = np.empty(n)
    ones.fill(1)
    centre_arr = np.dot(centre.reshape(centre.shape[0], 1),
        ones.reshape(1, ones.shape[0])).transpose()
    v = relative_vertices - centre_arr

    # find the magnitude of each of the vertices
    distance = np.sqrt(np.sum(v**2, axis=1))

    # get indices of values between radius and triangles radius
    relative_indices = np.where((distance > radius + mutil.epsilon) &
        (distance < np.sqrt(4 * pow(max_radius, 2) - pow(radius, 2))
        + mutil.epsilon))[0]

    maxv = False

    if relative_indices.size == 0:
        maxv = True
    else:
        # indices in sphere
        insphere_indices = np.where(distance <= radius + mutil.epsilon)

        maxv = True

        # loop through each index in relative indices
        for r_index in np.nditer(relative_indices):
            # create the new others list
            others = np.setdiff1d(relative[insphere_indices], basis)
            others = np.append(others, relative[r_index])

            # calculate minsphere through points
            c, r, b = sphere.minsphere(vertices, np.array([]), basis, others,
                max_radius)

            # not max if minsphere is smaller than max radius
            if r < (max_radius - mutil.epsilon):
                maxv = False
                break

    return maxv

def enumerate(vertices, fixed, others, explored, epsilon):
    """ Creates a list of all maximal cells of the cech complex

    Args:
        vertices: vertex list
        fixed: list of indices of vertices that have already been fixed
        others: vertices yet to be considered
        explored: cells that have been explored
        epsilon: radius

    Returns:
        centre: list of centre of cells
        radius: list of cell radi
        basis: list of cell basis
    """
    # get the sizes of the narrays
    d = vertices.shape[1]
    n = vertices.shape[0]

    k = fixed.shape[0]
    m = others.shape[0]

    if k>0:
        # use sphere through already fixed vertices
        c, r = sphere.through(vertices, fixed.astype(int))
    else:
        # use sphere centered through average points in others
        c = np.mean(vertices[others-1,:], 0)
        r = -np.inf

    # lists of cells
    centre_list = np.empty([0,3])
    radius_list = np.empty([0])
    basis_list = np.empty([0, d - k])

    if r < epsilon + mutil.epsilon:
        if (k==d+1) or (m==0):
            if k>0:
                ones = np.empty(explored.size)
                ones.fill(1)

                # translate vertices so that c is at the origin
                v = vertices[explored-1, :] - np.dot(
                    c.reshape(c.shape[0], 1),
                    ones.reshape(1, ones.shape[0])).transpose()

                # find the magnitude of each of the vertices
                distance = np.sqrt(np.sum(v**2, axis=1))

                # find all the vertices that are near sphere boundary
                boundary_vertices = np.where((distance <= r+mutil.epsilon) &
                    (distance >= r-mutil.epsilon))[0]

                if boundary_vertices.size == 0:
                    # no points on boundary so find new min bounding sphere
                    cp, rp, bs = sphere.minsphere(vertices, np.array([]),
                        np.array([fixed[0]]), fixed[1:k-1], epsilon)

                    if r<=rp+mutil.epsilon:
                        relative = np.append(others, explored)
                        if ismax(vertices, cp, rp, bs, relative, epsilon):
                            # add the sphere to lists
                            centre_list = np.r_[centre_list, cp.reshape(1, 3)]
                            radius_list = np.r_[radius_list, np.array([rp])]

                            new_bl = np.empty([1, d + 1 - k])
                            new_bl.fill(0)
                            new_bl[0, 0] = fixed
                            basis_list = new_bl

        else:
            new_explored = np.append(others, explored)

            centre_list, radius_list, basis_list = enumerate(vertices, fixed,
                np.array([]).astype(int), new_explored.astype(int), epsilon)

            ones = np.empty(others.size)
            ones.fill(1)

            # translate vertices so that c is at the origin
            v = vertices[others-1, :] - (np.dot(c.reshape(c.shape[0], 1),
                ones.reshape(1, ones.shape[0])).transpose())

            # find the magnitude of each of the vertices
            distance = np.sqrt(np.sum(v**2, axis=1))

            # sort the distances descending
            sorted_d_i = np.argsort(distance)[::-1][:n]
            others = others[sorted_d_i]

            for i in range(0, m):
                new_fixed = np.append(fixed, others[i])
                new_others = others[i+1:m]
                new_explored = np.append(others[0:i], explored)

                c, r, b = enumerate(vertices, new_fixed, new_others,
                    new_explored.astype(int), epsilon)

                centre_list = np.r_[centre_list, c]
                radius_list = np.r_[radius_list, r]

                if b.shape[0] != 0:
                    basis_list = np.r_[basis_list, b]

    return centre_list, radius_list, basis_list

def recursive_enumerate(vertices, indices, epsilon, depth):
    """ Recursive approach to find the maximal cells of each cech complex

    Args:
        vertices: vertex list
        indices: indices of the vertices to consider
        epsilon: radius
        depth: depth of recursion

    Returns:
        centre: list of centre cells
        radius: list of cell radi
        basis: list of cell basis
    """
    # get the sizes of the vertices
    d = vertices.shape[1]
    n = vertices.shape[0]

    m = indices.shape[0]

    # lists of cells
    centre_list = np.empty([0,3])
    radius_list = np.empty([0])
    basis_list = np.empty([0])

    if m<=30:
        centre_list, radius_list, basis_list = enumerate(vertices,
            np.array([]), indices.astype(int)+1, np.array([]), epsilon)
    else:
        points = vertices[indices.astype(int),:].transpose()
        a = np.cov(points)
        eigd, eigv = la.eig(a)

        index = np.argmax(eigd)
        evector = eigv[:,index]

        proj = np.dot(points.transpose(),
            evector.reshape([3, 1])).transpose()[0]

        if np.mod(proj.shape[0], 2) == 0:
            mpt = np.median(proj)
        else:
            mpt = np.median(np.r_[proj, np.inf])

        ileft = np.nonzero(proj <= mpt + epsilon + mutil.epsilon)[0].astype(int)
        iright = np.nonzero(proj > mpt - epsilon - mutil.epsilon)[0].astype(int)

        if np.maximum(ileft.shape[0], iright.shape[0]) > (0.8*m):
            centre_list, radius_list, basis_list = enumerate(vertices,
                np.array([]), indices.astype(int)+1, np.array([]), epsilon)
        else:
            c1, r1, b1 = recursive_enumerate(vertices, indices[ileft],
                epsilon, depth+1)
            dleft = np.dot(c1, evector)

            i1 = np.nonzero(dleft <= mpt + mutil.epsilon)[0].astype(int)
            c2, r2, b2 = recursive_enumerate(vertices, indices[iright],
                epsilon, depth+1)
            dright = np.dot(c2, evector)

            i2 = np.nonzero(dright > mpt - mutil.epsilon)[0].astype(int)

            centre_list = np.r_[c1[i1, :], c2[i2, :]]
            radius_list = np.r_[r1[i1], r2[i2]]
            basis_list = np.r_[b1[i1, :], b2[i2, :]]

    return centre_list, radius_list, basis_list
