from flask import Flask, jsonify, request, url_for
from flask.ext.api import status
import numpy as np
import pandas as pd

from context import off_file
from context import tasks

app = Flask(__name__)

ALLOWED_VERTEX_EXT = set(['off'])
ALLOWED_CSV_EXT = set(['csv'])

def allowed_file(filename, allowed):
    return '.' in filename and filename.rsplit('.', 1)[1] in allowed

@app.route('/recenum', methods=['POST'])
def run_recenum():
    # get the form values
    epsilon = float(request.form['epsilon'])
    depth = int(request.form['depth'])

    # get the file
    f = request.files['vertices']

    # check the file has been uploaded
    if not f:
        return 'Missing file', 400

    # check that the file is the correct format
    if not allowed_file(f.filename, ALLOWED_VERTEX_EXT):
        return 'Only .off files supported', 415

    # read file into numpy array
    vertices = off_file.read_off_panda(f)

    # run the method
    task = tasks.recursive_enumerate.delay(vertices, epsilon, depth)

    # return immediate response code stating that it's being processed
    return jsonify({}), 202, {'Location': url_for('recenum_status', task_id=task.id)}

@app.route('/cech', methods=['POST'])
def run_cech():
    # get the form values
    epsilon = float(request.form['epsilon'])

    # get the file
    f = request.files['vertices']

    # check that the file has been uploaded
    if not f:
        return 'Missing file', 400

    # check that the file is the correct format
    if not allowed_file(f.filename, ALLOWED_VERTEX_EXT):
        return 'Only .off files supported', 415

    # read file into numpy array
    vertices = off_file.read_off_panda(f)

    # run the method
    task = tasks.cech_task.delay(vertices, epsilon)

    # return immediate response code stating that it's being processed
    return jsonify({}), 202, {'Location': url_for('cech_status', task_id=task.id)}

@app.route('/entropy', methods=['POST'])
def run_entropy():
    # get the files
    a_file = request.files['a']
    p_file = request.files['p']

    # check that both files have been uploaded
    if not a_file:
        return 'Missing the a csv file', 400
    if not p_file:
        return 'Missing the p csv file', 400

    # check that files are the correct format
    if not allowed_file(a_file.filename, ALLOWED_CSV_EXT) or not allowed_file(
        p_file.filename, ALLOWED_CSV_EXT):
        return 'Only csv files supported', 415

    # read the files into numpy arrays
    a = pd.read_csv(a_file, delimiter=",", header=None).values
    p = pd.read_csv(p_file, delimiter=",", header=None).values

    # run the method
    task = tasks.simplical_complex_entropy.delay(a, p)

    # return immediate response code
    return jsonify({}), 202, {'Location': url_for('entropy_status', task_id=task.id)}

def task_response(task):
    if task.state == 'PENDING':
        response = {
            'state': 'Pending'
        }
    elif task.state != 'FAILURE':
        response = {
            'state': 'Completed',
            'results': task.info['results']
        }
    else:
        response = {
            'state': task.state,
            'status': str(task.info)
        }

    return response

@app.route('/recenum_status/<task_id>')
def recenum_status(task_id):
    task = tasks.recursive_enumerate.AsyncResult(task_id)
    response = task_response(task)
    return jsonify(response)

@app.route('/cech_satus/<task_id>')
def cech_status(task_id):
    task = tasks.cech_task.AsyncResult(task_id)
    response = task_response(task)
    return jsonify(response)

@app.route('/entropy_status/<task_id>')
def entropy_status(task_id):
    task = tasks.simplical_complex_entropy.AsyncResult(task_id)
    response = task_response(task)
    return jsonify(response)

app.run(debug=True)
